/* 
   signal-transfer mit USR1=0bit USR2=1bit zwischen programmen! 
*/

#ifndef CONFIG_SIG
#define CONFIG_SIG 1	/* if your system does not support signal, set to 0 */
#endif

#include <stdio.h>
#include <time.h>	/* definiert oft ulong (deshalb ULong) */
#include "error.h"
#include <unistd.h>
#include "sighand.h"


/* export: int sig1; int ini_handler(); */

volatile int sig1;	/* set by signal handler */
int stkptr=0;		/* stack mit stk[stkptr].p="pg-name",.i=counter */
struct struct_stk stk[20];	/* max. Tiefe stkptr=20 */


unsigned long starttime=0l;
long itime(){	/* ProzessZeit=clock()/CLOCKS_PER_SEC unixzeit=time */
  time_t i;time(&i);if(starttime==0l)starttime=i;return(long)(i-starttime);}
long iclock(){ return(long)clock();}


/*
+++TeX+++
	Dieser Programmteil ermoeglicht es dem Nutzer, mit dem laufenden
	Programm zu kommunizieren. Es kann zum Beispiel sehr praktisch
	sein zu erfahren, wo das Programm gerade rechnet.
	
	Die Kommunikation wird hier ueber Signale abgewickelt. 
	In UNIX-Systemen kann mit Hilfe des {\bf kill}-Befehls
	ein Signal an den Prozess gesendet werden. Das kann zum Beispiel
	ein STOP-Signal sein, es stehen aber auch zwei USER-Signale zur
	Verfuegung.
	
	Ist die Prozess-Identifikations-Nummer (PID) zum Beispiel 1234, so
	kann mit {\bf kill -SIGUSR1 1234} oder {\bf kill -SIGUSR2 1234}
	eines der beiden Signale an den Prozess gesendet werden.
	Das Programm reagiert dann beispielsweise mit der Ausgabe
	einer Tabelle, die die gerade ablaufenden Funktionen und einen
	spezifischen Zaehlerstand ausgibt.
	
	Die Tabelle kann dann bei Start eines Unterprogramms erweitert
	und bei verlassen des Programms wieder verkleinert werden.
	Mit einer Kombination von $n$ Signalen SIGUSR2 und einem
	SIGUSR1 koennen auch kompliziertere Reaktionen vom Programm 
	erzeugt werden, da das Programm die Variable {\bf sig1}, die
	dann $n$ enthaelt, jederzeit auslesen kann.
---TeX---
*/

void o_stack(){ int i;
   err_pg(0);	/* err-status */
   for(i=0;i<stkptr;i++){
     printf("#%d# %20s",i,stk[i].p); if(stk[i].i)printf(" %lu",*(stk[i].i));
     printf("\n");
   }
}

#if CONFIG_SIG
#include <signal.h>
#define PID_FILE "status"
void my_handler(int sig){ static int x=0; 
    switch(sig){
      case SIGUSR1: switch(x){
                     case 0: if(!sig1) o_stack(); break;
                     case 1: /*  */ break;} 
                    sig1=x; x=0; break;
      case SIGUSR2: x++; break;
      case SIGALRM: /* alarm(60); */ break; 
      default:      sig1=sig;printf("\nsignal(%d) received ",sig);}
    signal(sig,my_handler); /* neu-umlenken */}

/* mehrfachaufruf moeglich! */
int ini_handler(int vvv){ /* better open and close_handler ??? */
   FILE *f1;
   /* store the own PID (Process IDentification number) in a file */
   /* can also be used, too communicate with the process (sh status) */
   f1=fopen(PID_FILE,"w");
   if(!f1)fprintf(stderr," error "__FILE__" %d: create "PID_FILE,__LINE__);
   if(f1){ if(vvv)printf(" write "PID_FILE);
     fprintf(f1,"#!sh\n# get status of spin-pg (automaticly generated)\n"
                "# use: sh "PID_FILE"\nkill -USR1 %ld # or kill -SIGUSR1\n"
                "# writes to stdout (redirected to nohup.out or whatever)\n",
                (long)getpid());
     fclose(f1); }
   sig1=0;
   signal(SIGUSR1,my_handler);	/* fkt(x); x=0; (PG-Status) */
   signal(SIGUSR2,my_handler);  /* x=x+1; */
   signal(SIGALRM,my_handler);  /*  */
   /* if(vvv&16)alarm(60); */	/* SIgnal SIGALRM in 60s */
   return 0;}

#else
int ini_handler(int vvv){ return 0; }	/* dummy funktion */
#endif	
