/* this works on SGI-c89,CC, DEC-c89 ...
 *  => OpenMP http://www.openmp.org/index.cgi?specs
 * pragma synchronize (if pfor in same parallel region or call more than once)
 *  ?????????????????
 */
#include <stdlib.h>
#include <stdio.h>
#define NumBlks  4   /* must be a 2^n number */
#define DimBlk   (16*1024*1024) /* mem=2*NumBlks*DimBlk*8B every process??? */
int mode=0;	/* 0=1CPUdummy, 1=1CPU, 2=4CPU, 3=4CPU(blocks interleaved) */

/* 133MHz-M/2   5.2s 35s/10 341s/100  CC -O0 -mp -64
   chall---1-M  16Mx1=72*1+16=88s  (dummy=16s (mp=5s))
   chall-mp2-M  16Mx1=20*4+16=94s  (vektor2 same blocks)
   chall-mp3-M  16Mx1=17*4+16=95s  (vector2 diff blocks)
   chall-mp4-M  16Mx1=93*4+16=353s (vector2 mix blocks)
 */


 /* a[i]=H[i][j]*b[j] */
void call_mult(int i, int j, double *b[NumBlks], double a[]){
  int k;
  for (k=0;k<DimBlk;k++){
    /* do some calculation */
    a[k]=b[((mode==4)?((j+k)&(NumBlks-1)):(j+j)>>2)][k^((DimBlk-1) | (i+j))];
  }
}

int blockmult(){
  double *a[NumBlks],*b[NumBlks],c;
  int i,j,jj,k,nb=NumBlks,ncpu=NumBlks;
  if(mode<2) ncpu=1;
  printf(" malloc %d * %e\n",NumBlks,(double)DimBlk);
  for (i=0;i<NumBlks;i++) a[i]=(double *)malloc( DimBlk * sizeof(double) );
  for (i=0;i<NumBlks;i++) b[i]=(double *)malloc( DimBlk * sizeof(double) );
  for (i=0;i<NumBlks;i++) if ( b[i]==NULL || a[i]==NULL ) exit(1);
  printf(" start calculation ...\n");
  for (i=0;i<NumBlks;i++) for (j=0;j<DimBlk;j++)  b[i][j]=j;
  for(k=0;k<1;k++)		/* repeat for getting more time */
  {
#pragma parallel byvalue(nb) shared(a,b) local(i,j,jj)
#pragma numthreads (ncpu)
{
#pragma pfor iterate(i=0;nb;1)
    for (i=0;i<NumBlks;i++) 
    {
      for (j=0;j<NumBlks;j++) 
      {
         jj=((mode==3)?(j+i)&(NumBlks-1):j); /* what is faster ? */
         if(mode!=0) call_mult(i,jj,b,a[i]); /* a[i]=H[i][jj]*b[jj] */
      }
    }
}
  }
  c=0.0;
  for (i=0;i<NumBlks;i++) for (j=0;j<DimBlk;j++)  c+=a[i][j];
  for (i=0;i<NumBlks;i++) { free(a[i]); free(b[i]); }
  printf(" c=%e\n",c);
  return 0;
}

int testsum(){  int i,ii,j=6,l=0,k[6],sum=0,suml;
#pragma parallel byvalue(j,l) shared(k,sum) local(i,ii,suml)
#pragma numthreads (6)
   { suml=0;
#pragma pfor iterate(i=l;j;1)
     for(i=l;i<j;i++)
     { k[i]=0; for(ii=0;ii<10000000;ii++) k[i]++;
       suml=k[i]; printf("%d %d\n",i,suml);
     }
#pragma critical
     {  sum+=suml; }
   } printf("%d %d summed\n",i,sum); return 0;
}

int main(int argc, char *argv[]){
  if(argc>1) { mode=atoi(argv[1]); printf(" mode=%d\n",mode); }
  if(argc<2) { printf(" options: [mode]\n mode: 0..3\n"); }
// testsum();
  blockmult();
  return 0;
}
